#include <string.h>
#include <stdlib.h>

#include <node.h>
#include <nan.h>

#include <bitprim/nodecint/chain/transaction_list.h>

#include <bitprim/js-api/chain/transaction_list.h>

namespace bitprim_ns {

using v8::FunctionCallbackInfo;
using v8::Isolate;
using v8::Local;
using v8::Handle;
using v8::Global;

using v8::Object;
using v8::String;
using v8::Boolean;
using v8::Value;
using v8::External;
using v8::Exception;
using v8::Number;
using v8::Persistent;
using v8::Function;
using v8::Uint8Array;
using v8::ArrayBuffer;
// using v8::Nan;


void bitprim_chain_transaction_list_construct_default(v8::FunctionCallbackInfo<v8::Value> const& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong number of arguments")));
        return;
    }

    transaction_list_t res = chain_transaction_list_construct_default();
    args.GetReturnValue().Set(External::New(isolate, res));
}

void bitprim_chain_transaction_list_destruct(v8::FunctionCallbackInfo<v8::Value> const& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong number of arguments")));
        return;
    }

    if ( ! args[0]->IsExternal()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    void* vptr = v8::External::Cast(*args[0])->Value();
    transaction_list_t transaction_list = (transaction_list_t)vptr;

    chain_transaction_list_destruct(transaction_list);
}

void bitprim_chain_transaction_list_push_back(v8::FunctionCallbackInfo<v8::Value> const& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 2) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong number of arguments")));
        return;
    }

    if ( ! args[0]->IsExternal()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    if ( ! args[1]->IsExternal()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }
    
    void* vptr = v8::External::Cast(*args[0])->Value();
    transaction_list_t transaction_list = (transaction_list_t)vptr;

    void* vptr2 = v8::External::Cast(*args[1])->Value();
    transaction_t transaction = (transaction_t)vptr2;

    chain_transaction_list_push_back(transaction_list, transaction);
}

void bitprim_chain_transaction_list_count(v8::FunctionCallbackInfo<v8::Value> const& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong number of arguments")));
        return;
    }

    if ( ! args[0]->IsExternal()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }
    
    void* vptr = v8::External::Cast(*args[0])->Value();
    transaction_list_t transaction_list = (transaction_list_t)vptr;

    uint64_t res = chain_transaction_list_count(transaction_list);
    args.GetReturnValue().Set(Number::New(isolate, res));
}

void bitprim_chain_transaction_list_nth(v8::FunctionCallbackInfo<v8::Value> const& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 2) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong number of arguments")));
        return;
    }

    if ( ! args[0]->IsExternal()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    if ( ! args[1]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }
    
    void* vptr = v8::External::Cast(*args[0])->Value();
    transaction_list_t transaction_list = (transaction_list_t)vptr;

    uint64_t n = args[1]->IntegerValue(Nan::GetCurrentContext()).FromJust();

    transaction_t res = chain_transaction_list_nth(transaction_list, n);
    args.GetReturnValue().Set(External::New(isolate, res));
}

}  // namespace bitprim_ns
